#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <ctype.h>
#include "tag.h"
#include "symboltable.h"


// Lexing functions
void lex_file(FILE *input, FILE *output);
void lex_line(const char *line, FILE *output);
int lex_token(Tag *dest, const char *line);

// Parsing functions
void parse_file(FILE *input, FILE *output);
void parse_class(Tag **current, Tag **lookahead, FILE *input, FILE *output);
void parse_class_var_dec(Tag **current, Tag **lookahead, FILE *input, FILE *output);
void parse_sub_dec(Tag **current, Tag **lookahead, FILE *input, FILE *output);
void parse_parameter_list(Tag **current, Tag **lookahead, FILE *input, FILE *output);
void parse_sub_body(Tag **current, Tag **lookahead, FILE *input, FILE *output);
void parse_var_dec(Tag **current, Tag **lookahead, FILE *input, FILE *output);
void parse_statements(Tag **current, Tag **lookahead, FILE *input, FILE *output);
void parse_let_statement(Tag **current, Tag **lookahead, FILE *input, FILE *output);
void parse_if_statement(Tag **current, Tag **lookahead, FILE *input, FILE *output);
void parse_while_statement(Tag **current, Tag **lookahead, FILE *input, FILE *output);
void parse_do_statement(Tag **current, Tag **lookahead, FILE *input, FILE *output);
void parse_return_statement(Tag **current, Tag **lookahead, FILE *input, FILE *output);
void parse_expression(Tag **current, Tag **lookahead, FILE *input, FILE *output);
void parse_sub_call(Tag **current, Tag **lookahead, FILE *input, FILE *output);
void parse_term(Tag **current, Tag **lookahead, FILE *input, FILE *output);
void parse_expression_list(Tag **current, Tag **lookahead, FILE *input, FILE *output);

// Holds pointers to all the data we'll need during the compilation process to save us from having to pass seven
// arguments around in every single function call.
struct CompileData {
    Tag *current;
    Tag *lookahead;
    SymbolTable *class_table;
    SymbolTable *subroutine_table;
    char *class_name;
    FILE *input;
    FILE *output;
}; typedef struct CompileData CompileData;

void compile_file(FILE *input, FILE *output);
void compile_class(CompileData *data);
void compile_class_var_dec(CompileData *data);
void compile_sub_dec(CompileData *data);
void compile_parameter_list(CompileData *data);
void compile_sub_body(CompileData *data, Keyword sub_kind, char *sub_name);
void compile_var_dec(CompileData *data);
void compile_statements(CompileData *data);
void compile_let_statement(CompileData *data);
void compile_if_statement(CompileData *data);
void compile_while_statement(CompileData *data);
void compile_do_statement(CompileData *data);
void compile_return_statement(CompileData *data);
void compile_expression(CompileData *data);
void compile_sub_call(CompileData *data);
void compile_term(CompileData *data);
int compile_expression_list(CompileData *data);

int main(int argc, char *argv[]) {
    if (argc != 3) {
        printf("Please supply two arguments: an input .jack file, and an output .vm file.");
        exit(EXIT_FAILURE);
    }
    char *input_name = argv[1];
    char *output_name = argv[2];

    FILE *input = fopen(input_name, "r");
    if (input == NULL) {
        printf("Error opening input file.\n");
        exit(EXIT_FAILURE);
    }
    FILE *lex_output = fopen("lex_out.xml", "w");
    if (lex_output == NULL) {
        printf("Error opening lex_out.xml for writing.\n");
        exit(EXIT_FAILURE);
    }
    lex_file(input, lex_output);
    fclose(input);
    fclose(lex_output);

    FILE *lex_input = fopen("lex_out.xml", "r");
    if (lex_input == NULL) {
        printf("Error opening lex_out.xml for reading.\n");
        exit(EXIT_FAILURE);
    }
    FILE *parse_output = fopen("parse_out.xml", "w");
    if (parse_output == NULL) {
        printf("Error opening parse_out.xml for writing.\n");
        exit(EXIT_FAILURE);
    }
    parse_file(lex_input, parse_output);
    fclose(lex_input);
    fclose(parse_output);

    FILE *parse_input = fopen("parse_out.xml", "r");
    if (parse_input == NULL) {
        printf("Error opening parse_out.xml for reading.\n");
        exit(EXIT_FAILURE);
    }
    FILE *output = fopen(output_name, "w");
    if (output == NULL) {
        printf("Error opening output file.");
        exit(EXIT_FAILURE);
    }

    // Debugging help
    setbuf(output, NULL);

    compile_file(parse_input, output);
    fclose(parse_input);
    fclose(output);

    return EXIT_SUCCESS;
}

// Outputs a tokenised version of input to output.
void lex_file(FILE *input, FILE *output) {
    char line[MAX_LINE_LENGTH];
    write_non_terminal(NT_TOKENS, false, output);
    while (fgets(line, MAX_LINE_LENGTH, input) != NULL) {
        lex_line(line, output);
    }
    write_non_terminal(NT_TOKENS, true, output);
}

// Reads the next line from input, tokenises it, and writes the resulting tokens to output.
void lex_line(const char *line, FILE *output) {
    // This will be true if we are currently inside a comment of the form /*...*/ (which may have started on a
    // previous line).
    static bool inside_comment = false;

    int pos = 0;
    // In each iteration of this loop, everything up to line[pos] has been lexed.
    // NB newlines aren't tokens in Jack, so no need to care about \n versus \r\n.
    while (line[pos] != '\n' && line[pos] != '\0') {
        // Ignore all whitespace.
        if (line[pos] == ' ' || line[pos] == '\t') {
            pos++;
            continue;
        }

        // If we are inside a comment, ignore everything until the end of the comment.
        if (inside_comment) {
            if (line[pos] == '*' && line[pos+1] == '/') {
                pos++;
                inside_comment = false;
            }
            pos++;
            continue;
        }

        // Ignore all comments.
        if (line[pos] == '/' && line[pos+1] == '/') {
            return;
        } else if (line[pos] == '/' && line[pos+1] == '*') {
            // Note the comment may end on this line
            inside_comment = true;
            pos += 2;
            continue;
        }

        // Otherwise, we have at least one token, so lex it.
        Tag *next = malloc_tag();
        pos += lex_token(next, line + pos);
        write_tag(next, output);
        free_tag(&next);
    }
}

// Reads the next token from a non-empty, non-label, non-comment line into dest, then returns the number of characters
// in that token.
int lex_token(Tag *dest, const char *line) {
    int length;

    if (line[0] >= '0' && line[0] <= '9') {
        // Identifiers can't start with integers, so this must be an integer literal
        dest->type = INTEGER_LITERAL;
        char literal[MAX_LINE_LENGTH];
        for (length = 0; line[length] >= '0' && line[length] <= '9'; length++) {
            literal[length] = line[length];
        }
        literal[length] = '\0';
        dest->value.int_val = strtol(literal, NULL, 10);
    } else if (line[0] == '\"') {
        // Identifiers can't start with quotes, so this must be a string literal
        dest->type = STRING_LITERAL;
        for (length = 1; line[length] != '\"'; length++);
        length++;
        dest->value.str_val = (char *) malloc(length-1);
        strncpy(dest->value.str_val, line+1, length-2);
        dest->value.str_val[length-2] = '\0'; // Strncpy doesn't null-terminate.
    } else if (line[0] == '{' || line[0] == '}' || line[0] == '(' || line[0] == ')' || line[0] == '[' ||
               line[0] == ']' || line[0] == '.' || line[0] == ',' || line[0] == ';' || line[0] == '+' ||
               line[0] == '-' || line[0] == '*' || line[0] == '/' || line[0] == '&' || line[0] == '|' ||
               line[0] == '<' || line[0] == '=' || line[0] == '>' || line[0] == '~'){
        dest->type = SYMBOL;
        length = 1;
        dest->value.str_val = (char *)(malloc(2));
        dest->value.str_val[0] = line[0];
        dest->value.str_val[1] = '\0';
    } else { // We either have an identifier or a keyword.
        // Either way, it keeps going until a character other than a letter, a digit, or an underscore.
        length = 0;
        while(isalpha(line[length]) || isdigit(line[length]) || line[length] == '_'){
            length++;
        }

        dest->type = KEYWORD; // Default
        if (strncmp(line, "class", length) == 0 && length == 5) {
            dest->value.key_val = KW_CLASS;
        } else if (strncmp(line, "constructor", length) == 0 && length == 11) {
            dest->value.key_val = KW_CONSTRUCTOR;
        } else if (strncmp(line, "function", length) == 0 && length == 8) {
            dest->value.key_val = KW_FUNCTION;
        } else if (strncmp(line, "method", length) == 0 && length == 6) {
            dest->value.key_val = KW_METHOD;
        } else if (strncmp(line, "field", length) == 0 && length == 5) {
            dest->value.key_val = KW_FIELD;
        } else if (strncmp(line, "static", length) == 0 && length == 6) {
            dest->value.key_val = KW_STATIC;
        } else if (strncmp(line, "var", length) == 0 && length == 3) {
            dest->value.key_val = KW_VAR;
        } else if (strncmp(line, "int", length) == 0 && length == 3) {
            dest->value.key_val = KW_INT;
        } else if (strncmp(line, "char", length) == 0 && length == 4) {
            dest->value.key_val = KW_CHAR;
        } else if (strncmp(line, "boolean", length) == 0 && length == 7) {
            dest->value.key_val = KW_BOOLEAN;
        } else if (strncmp(line, "void", length) == 0 && length == 4) {
            dest->value.key_val = KW_VOID;
        } else if (strncmp(line, "true", length) == 0 && length == 4) {
            dest->value.key_val = KW_TRUE;
        } else if (strncmp(line, "false", length) == 0 && length == 5) {
            dest->value.key_val = KW_FALSE;
        } else if (strncmp(line, "null", length) == 0 && length == 4) {
            dest->value.key_val = KW_NULL;
        } else if (strncmp(line, "this", length) == 0 && length == 4) {
            dest->value.key_val = KW_THIS;
        } else if (strncmp(line, "let", length) == 0 && length == 3) {
            dest->value.key_val = KW_LET;
        } else if (strncmp(line, "do", length) == 0 && length == 2) {
            dest->value.key_val = KW_DO;
        } else if (strncmp(line, "if", length) == 0 && length == 2) {
            dest->value.key_val = KW_IF;
        } else if (strncmp(line, "else", length) == 0 && length == 4) {
            dest->value.key_val = KW_ELSE;
        } else if (strncmp(line, "while", length) == 0 && length == 5) {
            dest->value.key_val = KW_WHILE;
        } else if (strncmp(line, "return", length) == 0 && length == 6) {
            dest->value.key_val = KW_RETURN;
        } else { // We've now ruled out all the keywords, so it must be an identifier.
            dest->type = IDENTIFIER;
            dest->value.str_val = (char *)malloc(length+1);
            strncpy(dest->value.str_val, line, length);
            dest->value.str_val[length] = '\0'; // Strncpy doesn't null-terminate.
        }
    }
    return length;
}

// Input should be a tokenised file. Writes a parse tree in XML form to output.
void parse_file(FILE *input, FILE *output) {
    Tag **lookahead = malloc(sizeof(Tag *));
    *lookahead = NULL;
    Tag **current = malloc(sizeof(Tag *));
    *current = NULL;
    advance_tag(current, lookahead, input);
    advance_tag(current, lookahead, input);
    // Now *current points to the first tag in the file, and *lookahead points to the second tag.
    // Note that copy_tag and advance_tag handle malloc'ing and free'ing for us.

    // This tag is in there purely to conform to the XML standard --- see the lab sheet.
    if ((*current)->type != NON_TERMINAL || (*current)->value.nt_val != NT_TOKENS) {
        printf("Error: Token list should start with <tokens>.\n");
        exit(EXIT_FAILURE);
    }
    advance_tag(current, lookahead, input);

    // Now we should be at the class declaration, which we parse in full.
    if ((*current)->type != KEYWORD || (*current)->value.nt_val != KW_CLASS) {
        printf("Error: All Jack files should consist of a single class.\n");
        exit(EXIT_FAILURE);
    }
    parse_class(current, lookahead, input, output);

    // Now our current tag should be the </tokens> tag at the end of the file, and lookahead should be null.
    if ((*lookahead) != NULL) {
        printf("Error: All Jack files should consist of a single class.\n");
        exit(EXIT_FAILURE);
    } if ((*current)->type != NON_TERMINAL || (*current)->value.nt_val != NT_TOKENS) {
        printf("Error: Token list should end with </tokens>.\n");
        exit(EXIT_FAILURE);
    }
    advance_tag(current, lookahead, input);
    free(current);
    free(lookahead);
}

// The general "contract" of a parse_*** function should be: If e.g. parse_if_statement is called, then "current" points
// to the first token of an <ifStatement> non-terminal, and "lookahead" points to the token after "current". On function
// return, "current" will point to the token after the last token in that <ifStatement> (i.e. after the ';') and
// "lookahead" will once again point to the token after "current".
// This is what careful error-checking looks like.
void parse_class(Tag **current, Tag **lookahead, FILE *input, FILE *output) {
    // Syntax is 'class', identifier, '{', {<classVarDec>}, {<subroutineDec>}, '}'
    write_non_terminal(NT_CLASS, false, output);
    advance_tag(current, lookahead, input);

    if ((*current)->type != IDENTIFIER) {
        printf("Error: <class> expected identifier token.\n");
        exit(EXIT_FAILURE);
    }
    copy_tag(current, lookahead, input, output);

    if ((*current)->type != SYMBOL || strcmp((*current)->value.str_val, "{") != 0) {
        printf("Error: <class> expected '{' token.\n");
        exit(EXIT_FAILURE);
    }
    advance_tag(current, lookahead, input);

    bool var_dec_over = false;
    while (true) {
        // If we see a } then the class is over, close the tag and return.
        if ((*current)->type == SYMBOL && strcmp((*current)->value.str_val, "}") == 0) {
            advance_tag(current, lookahead, input);
            write_non_terminal(NT_CLASS, true, output);
            return;
        }

        // A 'static' or a 'field' means this is a variable declaration. A 'constructor', 'function' or 'method' means
        // this is a subroutine declaration. Anything else is unexpected.
        if ((*current)->type == KEYWORD && ((*current)->value.key_val == KW_STATIC || (*current)->value.key_val == KW_FIELD)) {
            if (var_dec_over) {
                printf("Error: All variables must occur at start of class declaration.\n");
                exit(EXIT_FAILURE);
            }
            parse_class_var_dec(current, lookahead, input, output);
        } else if ((*current)->type == KEYWORD && ((*current)->value.key_val == KW_CONSTRUCTOR ||
                        (*current)->value.key_val == KW_METHOD || (*current)->value.key_val == KW_FUNCTION)) {
            var_dec_over = true;
            parse_sub_dec(current, lookahead, input, output);
        } else {
            printf("Error: Unexpected token in <class>.\n");
            exit(EXIT_FAILURE);
        }

        if (current == NULL) {
            printf("Error: <class> expected '}' token.\n");
            exit(EXIT_FAILURE);
        }
    }
}

// This is what a lack of error-checking looks like. It's much easier!
void parse_class_var_dec(Tag **current, Tag **lookahead, FILE *input, FILE *output) {
    // Syntax: ('static' | 'field'), ('int' | 'char' | 'boolean' | identifier), identifier, {',', identifier}*, ';'
    write_non_terminal(NT_CLASS_VAR_DEC, false, output);

    copy_tag(current, lookahead, input, output);
    copy_tag(current, lookahead, input, output);
    copy_tag(current, lookahead, input, output);

    while ((*current)->type != SYMBOL || strcmp((*current)->value.str_val, ";") != 0) {
        advance_tag(current, lookahead, input);
        copy_tag(current, lookahead, input, output);
    }
    advance_tag(current, lookahead, input);

    write_non_terminal(NT_CLASS_VAR_DEC, true, output);
}

void parse_sub_dec(Tag **current, Tag **lookahead, FILE *input, FILE *output) {
    // Syntax: ('constructor' | 'method' | 'function'), ('int' | 'char' | 'boolean' | 'void' | identifier), identifier,
    // '(', <parameterList>, ')', <subroutineBody>.
    // TODO: Start work here!
}

// Children in AST: <varDec>, <statements>.
void parse_sub_body(Tag **current, Tag **lookahead, FILE *input, FILE *output) {
    // Syntax: '{', {varDec}, <statements>, '}'
    // TODO
}

// Children in AST: variable keyword, type keyword, list of variable names.
void parse_var_dec(Tag **current, Tag **lookahead, FILE *input, FILE *output) {
    // Syntax: 'var', ('int' | 'char' | 'boolean' | identifier), identifier, {',', identifier}, ';'
    // TODO
}

// Children in AST: Only the statement type.
void parse_statements(Tag **current, Tag **lookahead, FILE *input, FILE *output) {
    // Syntax: {<letStatement> | <ifStatement> | <whileStatement> | <doStatement> | <returnStatement>}
    // TODO
}

// Children in AST: destination variable, [<expression>] if present, right-hand side <expression>.
void parse_let_statement(Tag **current, Tag **lookahead, FILE *input, FILE *output) {
    // Syntax: 'let', identifier, ['[', <expression>, ']'], '=', <expression>, ';'
    // TODO
}

// Children in AST: Condition <expression>, <statements> in if clause, <statements> in else clause (if present)
void parse_if_statement(Tag **current, Tag **lookahead, FILE *input, FILE *output) {
    // Syntax: 'if', '(', <expression>, ')', '{', <statements>, '}', ['else', '{', <statements>, '}']
    // TODO
}

// Children in AST: Condition <expression>, <statements> in loop body.
void parse_while_statement(Tag **current, Tag **lookahead, FILE *input, FILE *output) {
    // Syntax: 'while', '(', <expression>, ')', '{', <statements>, '}'
    // TODO
}

// Children in AST: Just the subroutine call.
void parse_do_statement(Tag **current, Tag **lookahead, FILE *input, FILE *output) {
    // Syntax: 'do', <expression>, ';'
    // TODO
}

// Children in AST: <expression> to return (if present).
void parse_return_statement(Tag **current, Tag **lookahead, FILE *input, FILE *output) {
    // Syntax: 'return', [<expression>], ';'
    // TODO
}

// Children in AST: As in original, <term>, {<op>, <term>}.
void parse_expression(Tag **current, Tag **lookahead, FILE *input, FILE *output) {
    // Syntax: <term>, {<op>, <term>}
    // where <op> ::= '+' | '-' | '*' | '/' | '&' | '|' | '<' | '>' | '='
    // TODO
}

// Children in AST: Everything but the brackets.
void parse_sub_call(Tag **current, Tag **lookahead, FILE *input, FILE *output) {
    // Syntax: identifier, ['.', identifier], '(', <expressionList>, ')'
    // TODO
}

// Children in AST: Everything except ()s around an <expression>.
void parse_term(Tag **current, Tag **lookahead, FILE *input, FILE *output) {
    // Syntax: integerLiteral | stringLiteral | 'true' | 'false' | 'null' | 'this' |
    //         identifier, ['[', <expression>, ']'] |
    //         '(', <expression>, ')' |
    //         ('-' | '~'), <term> |
    //         <subroutineCall>
    // TODO
}

// Children in AST: Just the <expression>s.
void parse_expression_list(Tag **current, Tag **lookahead, FILE *input, FILE *output) {
    // Syntax: [<expression>, {',', <expression>}]
    // TODO
}

void compile_file(FILE *input, FILE *output) {
    CompileData *data = malloc(sizeof(CompileData));
    data->input = input;
    data->output = output;
    data->class_name = NULL;

    data->class_table = NULL;
    data->subroutine_table = NULL;

    data->current = NULL;
    data->lookahead = NULL;

    advance_tag(&(data->current), &(data->lookahead), input);
    advance_tag(&(data->current), &(data->lookahead), input);

    compile_class(data);

    // *Lookahead, *current, *class and *subroutine will be NULL again at this point, no need to free them.
    free(data);
}

// Outputs all VM code for the given <class> with no side effects.
void compile_class(CompileData *data) {
    // AST format: class name identifier, <classVarDec>s, <subroutineDec>s.
    // Skip open tag.
    advance_tag(&(data->current), &(data->lookahead), data->input);

    // Store class name.
    char *current_class_name = data->current->value.str_val;
    data->class_name = malloc(strlen(current_class_name)+1);
    strcpy(data->class_name, current_class_name);
    advance_tag(&(data->current), &(data->lookahead), data->input);

    // Create class_table symbol table from <classVarDecs>.
    data->class_table = malloc_table();
    while (data->current->value.nt_val == NT_CLASS_VAR_DEC) {
        compile_class_var_dec(data);
    }

    // Compile all subroutines using class_table symbol table.
    while (data->current->value.nt_val == NT_SUBROUTINE_DEC) {
        compile_sub_dec(data);
    }

    // Cleanup, skip closing tag.
	free(data->class_name);
    free_table(data->class_table);
    advance_tag(&(data->current), &(data->lookahead), data->input);
}

// Adds the variables from the given <classVarDec> to the class_table symbol table.
void compile_class_var_dec(CompileData *data) {
    // TODO
}

// Outputs all VM code for the current <subroutineDec> with no side effects.
void compile_sub_dec(CompileData *data) {
    // TODO
}

// Outputs all VM code for the current <subBody>, populating the subroutine's symbol table in the process.
void compile_sub_body(CompileData *data, Keyword sub_kind, char *sub_name) {
    // TODO
}

// Add variable(s) from given <varDec> to subroutine symbol table.
void compile_var_dec(CompileData *data) {
    // TODO
}

// Outputs all VM code for the current <statements> with no side effects.
void compile_statements(CompileData *data) {
    // TODO
}

// Outputs all VM code for the current <letStatement> with no side effects.
void compile_let_statement(CompileData *data) {
    // TODO
}

// Outputs all VM code for the current <ifStatement> with no side effects.
void compile_if_statement(CompileData *data) {
    // TODO
}

// Outputs all VM code for the current <whileStatement> with no side effects.
void compile_while_statement(CompileData *data) {
    // TODO
}

// Outputs all VM code for the current <doStatement> with no side effects.
void compile_do_statement(CompileData *data) {
    // TODO
}

// Outputs all VM code for the current <returnStatement> with no side effects.
void compile_return_statement(CompileData *data) {
    // TODO
}

// Outputs VM code that evaluates the current <expression> and leaves its value on top of the stack.
void compile_expression(CompileData *data) {
    // TODO
}

// Outputs VM code that calls the given subroutine and leaves the result on the stack.
void compile_sub_call(CompileData *data) {
    // TODO
}

// Outputs VM code that evaluates each expression in the given <expressionList>, leaving the results on the stack, and
// returns the total number of expressions.
int compile_expression_list(CompileData *data) {
    // TODO
}

// Outputs VM code that evaluates the current <term> and leaves its value on top of the stack.
void compile_term(CompileData *data) {
    // TODO
}

// Add all given arguments to subroutine symbol table.
void compile_parameter_list(CompileData *data) {
	// TODO
}